import sys
import numpy as np
from scipy.io import FortranFile
from astropy.io import fits
import argparse
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.special import eval_legendre
from scipy.integrate import simps

def fits_to_ascii(
  input_filename, output_filename, cosmology,
  is_random=False, equal_weights=False, zrange=None
):
  # open fits file
  with fits.open(input_filename) as hdul:
    cat = hdul[1].data

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat['Z'] > zmin) & (cat['Z'] < zmax)
    cat = cat[ind]

  # convert redshifts to distances
  dist = cosmology.ComovingDistance(cat['Z'])
  x = dist * np.cos(cat['DEC'] * np.pi / 180) * np.cos(cat['RA'] * np.pi / 180)
  y = dist * np.cos(cat['DEC'] * np.pi / 180) * np.sin(cat['RA'] * np.pi / 180)
  z = dist * np.sin(cat['DEC'] * np.pi / 180)

  if not equal_weights:
    weight = cat['WEIGHT_FKP']
    if not is_random:
      weight *= cat['WEIGHT_SYSTOT'] * (cat['WEIGHT_CP'] + cat['WEIGHT_NOZ'] - 1)
  else:
    weight = np.ones(len(cat))

  #write result to output file
  cout = np.c_[x, y, z, weight]
  np.savetxt(output_filename, cout)

def fits_to_unformatted(
  input_filename, output_filename, cosmology,
  is_random=False, equal_weights=False, zrange=None
):
  # open fits file
  with fits.open(input_filename) as hdul:
    cat = hdul[1].data

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat['Z'] > zmin) & (cat['Z'] < zmax)
    cat = cat[ind]

  # convert redshifts to distances
  dist = cosmology.ComovingDistance(cat['Z'])
  x = dist * np.cos(cat['DEC'] * np.pi / 180) * np.cos(cat['RA'] * np.pi / 180)
  y = dist * np.cos(cat['DEC'] * np.pi / 180) * np.sin(cat['RA'] * np.pi / 180)
  z = dist * np.sin(cat['DEC'] * np.pi / 180)

  if not equal_weights:
    weight = cat['WEIGHT_FKP']
    if not is_random:
      weight *= cat['WEIGHT_SYSTOT'] * (cat['WEIGHT_CP'] + cat['WEIGHT_NOZ'] - 1)
  else:
    weight = np.ones(len(cat))

  #write result to output file
  cout = np.c_[x, y, z, weight]
  nrows, ncols = np.shape(cout)
  f = FortranFile(output_filename, 'w')
  nrows, ncols = np.shape(cout)
  f.write_record(nrows)
  f.write_record(ncols)
  f.write_record(cout)
  f.close()

def fits_to_cute(
  input_filename, output_filename, is_random=False,
  equal_weights=False, zrange=None
):
  # open fits file
  with fits.open(input_filename) as hdul:
    cat = hdul[1].data

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat['Z'] > zmin) & (cat['Z'] < zmax)
    cat = cat[ind] 

  ra = cat['RA']
  dec = cat['DEC']
  redshift = cat['Z']

  if not equal_weights:
    if is_random:
      weight = cat['WEIGHT_FKP']
    else:
      weight = cat['WEIGHT_FKP'] * cat['WEIGHT_SYSTOT']
  else:
    weight = np.ones(len(cat))

  #write result to output file
  cout = np.c_[ra, dec, redshift, weight]
  np.savetxt(output_filename, cout)


def revolver_to_unformatted(
  input_filename, output_filename, cosmology,
  equal_weights=False, zrange=None
):
  # open numpy file
  cat = np.load(input_filename)

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat[:,2] > zmin) & (cat[:,2] < zmax)
    cat = cat[ind] 

  # convert redshifts to distances
  dist = cosmology.ComovingDistance(cat[:,2])
  x = dist * np.cos(cat[:,1] * np.pi / 180) * np.cos(cat[:,0] * np.pi / 180)
  y = dist * np.cos(cat[:,1] * np.pi / 180) * np.sin(cat[:,0] * np.pi / 180)
  z = dist * np.sin(cat[:,1] * np.pi / 180)

  if not equal_weights:
    weight = cat[:,3]
  else:
    weight = np.ones(len(cat))

  #write result to output file
  cout = np.c_[x, y, z, weight]
  nrows, ncols = np.shape(cout)
  f = FortranFile(output_filename, 'w')
  nrows, ncols = np.shape(cout)
  f.write_record(nrows)
  f.write_record(ncols)
  f.write_record(cout)
  f.close()


def patchy_to_unformatted(
  input_filename, output_filename, cosmology,
  is_random=False, equal_weights=False, zrange=None
):
  # open text file
  cat = np.genfromtxt(input_filename)

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat[:,2] > zmin) & (cat[:,2] < zmax)
    cat = cat[ind]

  # convert redshifts to distances
  dist = cosmology.ComovingDistance(cat[:,2])
  x = dist * np.cos(cat[:,1] * np.pi / 180) * np.cos(cat[:,0] * np.pi / 180)
  y = dist * np.cos(cat[:,1] * np.pi / 180) * np.sin(cat[:,0] * np.pi / 180)
  z = dist * np.sin(cat[:,1] * np.pi / 180)

  if not equal_weights:
    if is_random:
      weight = (cat[:,5] * cat[:,6]) / (1 + 10000 * cat[:,3])
    else:
      weight = (cat[:,6] * cat[:,7]) / (1 + 10000 * cat[:,4])
  else:
    weight = np.ones(len(cat))

  #write result to output file
  cout = np.c_[x, y, z, weight]
  nrows, ncols = np.shape(cout)
  f = FortranFile(output_filename, 'w')
  nrows, ncols = np.shape(cout)
  f.write_record(nrows)
  f.write_record(ncols)
  f.write_record(cout)
  f.close()


def patchy_to_revolver(
  input_filename, output_filename,
  is_random=False, zrange=None
):
  # open text file
  cat = np.genfromtxt(input_filename)

  if zrange is not None:
    zmin, zmax = zrange
    ind = (cat[:,2] > zmin) & (cat[:,2] < zmax)
    cat = cat[ind]

  ra = cat[:,0]
  dec = cat[:,1]
  z = cat[:,2]

  if is_random:
    fkp = 1 / (1 + 10000 * cat[:,3])
    veto = cat[:,5]
    cp = cat[:,6]
  else:
    fkp = 1 / (1 + 10000 * cat[:,4])
    veto = cat[:,6]
    cp = cat[:,7]

  #write result to output file
  cout = np.c_[ra, dec, z, fkp, cp, veto]
  np.savetxt(output_filename, cout)

def read_array_2d(input_filename):
  data = np.genfromtxt(input_filename)
  dim1 = np.unique(data[:,0])
  dim2 = np.unique(data[:,1])

  vary_dim2 = False
  if data[0,0] == data[1,0]:
      vary_dim2 = True

  result = np.zeros([len(dim1), len(dim2)])
  counter = 0
  if vary_dim2:
      for i in range(len(dim1)):
          for j in range(len(dim2)):
              result[i, j] = data[counter, 2]
              counter += 1
  else:
      for i in range(len(dim2)):
          for j in range(len(dim1)):
              result[j, i] = data[counter, 2]
              counter += 1
  return dim1, dim2, result

def get_multipole(ell, s, mu, xi_smu):
  multipole = np.zeros(xi_smu.shape[0])
  if mu.min() < 0:
      factor = 2
      mumin = -1
  else:
      factor = 1
      mumin=0
  for i in range(xi_smu.shape[0]):
      mufunc = InterpolatedUnivariateSpline(mu, xi_smu[i, :], k=3, ext=3)
      xaxis = np.linspace(mumin, 1, 1000)
      lmu = eval_legendre(ell, xaxis)
      yaxis = mufunc(xaxis) * (2 * ell + 1) / factor * lmu
      multipole[i] = simps(yaxis, xaxis)
  return s, multipole


import sys
import numpy as np
from scipy.io import FortranFile

def ascii_to_unformatted(input_filename, output_filename,
  pos_cols=[0, 1, 2], vel_cols=None, weight_cols=None):
  # import data
  data = np.genfromtxt(input_filename)
  data = data.astype(np.float64)
  
  pos = data[:, pos_cols]
  cout = pos # default catalogue with only positions

  if vel_cols is not None:
    vel = data[:, vel_cols]
    cout = np.c_[cout, vel]
  if weight_cols is not None:
    weight = data[:, weight_cols]
    cout = np.c_[cout, weight]

  f = FortranFile(output_filename, 'w')
  nrows, ncols = np.shape(cout)
  f.write_record(nrows)
  f.write_record(ncols)
  f.write_record(cout)
  f.close()

def npy_to_unformatted(array, output_filename):
  array = array.astype(np.float64)
  f = FortranFile(output_filename, 'w')
  nrows, ncols = np.shape(array)
  f.write_record(nrows)
  f.write_record(ncols)
  f.write_record(array)
  f.close()

def read_array_2d(input_filename):
  data = np.genfromtxt(input_filename)
  dim1 = np.unique(data[:,0])
  dim2 = np.unique(data[:,1])

  vary_dim2 = False
  if data[0,0] == data[1,0]:
      vary_dim2 = True

  result = np.zeros([len(dim1), len(dim2)])
  counter = 0
  if vary_dim2:
      for i in range(len(dim1)):
          for j in range(len(dim2)):
              result[i, j] = data[counter, 2]
              counter += 1
  else:
      for i in range(len(dim2)):
          for j in range(len(dim1)):
              result[j, i] = data[counter, 2]
              counter += 1
  return dim1, dim2, result



import sys
from os import path
import numpy as np
import subprocess
from scipy.io import FortranFile

def generate_centres(
  ncentres, output_filename, sampling='uniform',
  sampling_filename=None, xmin=0, xmax=0,
  ymin=0, ymax=0, zmin=0,
  zmax=0, output_format='unformatted'
):
  np.random.seed(0)

  if sampling == 'subsampling':
    # check if file exists
    if not path.isfile(sampling_filename):
      raise FileNotFoundError(f'{sampling_filename} does not exist.')
    # check if this is a numpy file
    if '.npy' in sampling_filename:
      centres = np.load(sampling_filename)
    else:
      # if not, check if it is a text file
      try:
        centres = np.genfromtxt(sampling_filename)
      except:
        # else, check if it is an unformatted file
        try:
          fin = FortranFile(sampling_filename, 'r')
          nrows = fin.read_ints()[0]
          ncols = fin.read_ints()[0]
          pos = fin.read_reals(dtype=np.float64).reshape(nrows, ncols)
          idx = np.random.choice(nrows, size=ncentres, replace=False)
          centres = pos[idx]
        except:
          sys.exit('Format of sampling file not recognized.')

  elif sampling == 'uniform':
      x = np.random.uniform(xmin, xmax, ncentres)
      y = np.random.uniform(ymin, ymax, ncentres)
      z = np.random.uniform(zmin, zmax, ncentres)
      centres = np.c_[x, y, z]
  else:
      sys.exit('Sampling type not recognized')

  centres = centres.astype('float64')
  if output_format == 'unformatted':
    f = FortranFile(output_filename, 'w')
    nrows, ncols = np.shape(centres)
    f.write_record(nrows)
    f.write_record(ncols)
    f.write_record(centres)
    f.close()
  elif output_format == 'ascii':
    np.savetxt(output_filename, centres)
  elif output_format == 'numpy':
    np.save(output_filename, centres)
  else:
    sys.exit('Output format not recognized.')

  return centres
  

def filtered_density(
  centres_filename, tracers_filename, output_filename,
  filter_type, filter_size, ngrid, box_size,
  nthreads=1, dim1_min=0, dim1_max=None,
  output_format='unformatted'
):

  # check if files exist
  if not path.isfile(centres_filename):
    raise FileNotFoundError(f'{centres_filename} does not exist.')

  if not path.isfile(tracers_filename):
    raise FileNotFoundError(f'{tracers_filename} does not exist.')


  if dim1_max == None:
    if filter_type == 'tophat':
            dim1_max = filter_size
    elif filter_type == 'gaussian':
            dim1_max = 5 * filter_size

  binpath = path.join(path.dirname(__file__),
    'bin', '{}_filter.exe'.format(filter_type))

  cmd = [
    binpath, centres_filename, tracers_filename,
    output_filename, str(box_size), str(dim1_min),
    str(dim1_max), str(filter_size), str(ngrid),
    str(nthreads)
  ]

  subprocess.call(cmd)

  # open filter file
  f = FortranFile(output_filename, 'r')
  smoothed_delta = f.read_ints()[0]
  smoothed_delta = f.read_reals(dtype=np.float64)
  f.close()

  if output_format != 'unformatted':
    if output_format == 'npy':
      subprocess.call(['rm', output_filename])
      np.save(output_filename, smoothed_delta)
    elif output_format == 'ascii':
      np.savetxt(output_filename, smoothed_delta)
    else:
      print('Output format not recognized. Using unformatted F90 file.')
  
  return smoothed_delta


def split_centres(
  centres_filename, filter_filename, quantiles,
  handle=None, output_format='unformatted'
):

  # read centres
  # first check if this is a numpy file
  if '.npy' in centres_filename:
    centres = np.load(centres_filename)
  else:
    # if not, check if it is a text file
    try:
      centres = np.genfromtxt(centres_filename)
    except:
      # else, check if it is an unformatted file
      try:
        fin = FortranFile(centres_filename, 'r')
        nrows = fin.read_ints()[0]
        ncols = fin.read_ints()[0]
        centres = fin.read_reals(dtype=np.float64).reshape(nrows, ncols)
      except:
        sys.exit('Format of centres file not recognized.')

  # read smoothed densities
  # first check if this is a numpy file
  if '.npy' in filter_filename:
    smoothed_delta = np.load(filter_filename)
    ncentres = len(smoothed_delta)
  else:
    # if not, check if it is a text file
    try:
      smoothed_delta = np.genfromtxt(filter_filename)
      ncentres = len(smoothed_delta)
    except:
      # else, check if it is an unformatted file
      try:
        f = FortranFile(filter_filename, 'r')
        ncentres = f.read_ints()[0]
        smoothed_delta = f.read_reals(dtype=np.float64)
        f.close()
      except:
        sys.exit('Format of filter file not recognized.')
  idx = np.argsort(smoothed_delta)

  # sort centres using smoothed densities
  sorted_centres = centres[idx]

  # generate quantiles
  binned_centres = {}
  for i in range(1, quantiles + 1):
      binned_centres['DS{}'.format(i)] = sorted_centres[int((i-1)*ncentres/quantiles):int(i*ncentres/quantiles)]
      cout = binned_centres['DS{}'.format(i)]

      if handle != None:
            output_filename = handle + '_DS{}'.format(i)
      else:
          output_filename = centres_filename.split('.unf')[0] + '_DS{}'.format(i)
      
      if output_format == 'unformatted':
        output_filename += '.unf'
        f = FortranFile(output_filename, 'w')
        f.write_record(np.shape(cout)[0])
        f.write_record(np.shape(cout)[1])
        f.write_record(cout)
        f.close()
      
      elif output_format == 'ascii':
        output_filename += '.dat'
        np.savetxt(output_filename, cout)

      elif output_format == 'npy':
        np.save(output_filename, cout)
      
      else:
        sys.exit('Output format not recognized.')

  return binned_centres



import numpy as np
from scipy.integrate import quad, simps
from scipy.special import hyp2f1, legendre, eval_legendre
from scipy.interpolate import InterpolatedUnivariateSpline

class Cosmology:
    '''
    Class for cosmological calculations.
    '''
    def __init__(self,
                omega_m=0.308,
                h=0.676,
                c=299792.458):

        omega_lambda = 1.0 - omega_m
        H0 = 100.0

        # we generate a mapping between distance and redshift
        # over which we can interpolate later to speed up 
        # calculations
        zgrid = np.linspace(0, 4, 1000)
        rgrid = np.zeros_like(zgrid)
        for i in range(len(zgrid)):
            rgrid[i] = quad(lambda x: (c/H0) / np.sqrt(omega_m * (1 + x) ** 3
                      + omega_lambda), 0, zgrid[i])[0]

        self.H0 = H0
        self.h = h
        self.c = c
        self.omega_m = omega_m
        self.omega_lambda = omega_lambda
        self.zgrid = zgrid
        self.rgrid = rgrid

    def Ez(self, z):
        ez = np.sqrt(self.omega_m * (1 + z) ** 3 + self.omega_lambda)
        return ez

    def HubbleParameter(self, z):
        return self.H0 * self.Ez(z)

    # comoving distance in Mpc/h
    def ComovingDistance(self, z):
        return np.interp(z, self.zgrid, self.rgrid)

    # angular diameter distance in Mpc/h
    def AngularDiameterDistance(self, z):
        return np.interp(z, self.zgrid, self.rgrid) / (1 + z)

    # redshift at a given comoving distance
    def Redshift(self, r):
        return np.interp(r, self.rgrid, self.zgrid)

    # growth factor at a given redshift
    def GrowthFactor(self, z):
        az = 1. / (1 + z)
        growth = az ** 2.5 * np.sqrt(self.omega_lambda + self.omega_m * az ** (-3.)) * \
                hyp2f1(5. / 6, 3. / 2, 11. / 6, -(self.omega_lambda * az ** 3.) / self.omega_m) / \
                hyp2f1(5. / 6, 3. / 2, 11. / 6, -self.omega_lambda / self.omega_m)
        return growth

    # linear growth rate at a given redshift
    def GrowthRate(self, z):
        f = ((self.omega_m * (1 + z)**3.) / (self.omega_m * (1 + z)**3 + self.omega_lambda))**0.55
        return f